@@ -304,7 +304,10 @@ def trainmodel2(inifile):
             importances = list(clf.feature_importances_)
             log_df = generateFeatureImportanceLog(importances)
             generate_features_importance_log = 'yes'
-            N_feature_importance_bars = config.getint('create ensemble settings', 'N_feature_importance_bars')
+            try:
+                N_feature_importance_bars = config.getint('create ensemble settings', 'N_feature_importance_bars')
+            except ValueError:
+                print('SimBA could not find how many feature importance bars you want to generate. Double check that you entered an integer in the "Genereate Feature Importance Bar Graph" entry box"')
             print('Generating feature importance bar graph...')
             generateFeatureImportanceBarGraph(log_df, N_feature_importance_bars)
         if generate_features_importance_bar_graph != 'yes':
